@@ -29,8 +29,15 @@
  * time has been reached. It will then pass-through both audio and video,
  * starting from that specific timecode or running time, making sure that
  * audio starts as early as possible after the video (or at the same time as
- * the video). In the "audio-after-video" mode, it only drops audio buffers
- * until video has started.
+ * the video). In the "video-first" mode, it only drops audio buffers until
+ * video has started.
+ *
+ * The "recording" property acts essentially like a valve connected before
+ * everything else. If recording is FALSE, all buffers are dropped regardless
+ * of settings. If recording is TRUE, the other settings (mode,
+ * target-timecode, target-running-time, etc) are taken into account. Audio
+ * will always start and end together with the video, as long as the stream
+ * itself doesn't start too late or end too early.
  *
  * ## Example launch line
  * |[
@@ -85,6 +92,7 @@ enum
   PROP_TARGET_TIME_CODE_STRING,
   PROP_TARGET_RUNNING_TIME,
   PROP_END_TIME_CODE,
+  PROP_RECORDING,
   PROP_MODE
 };
 
@@ -92,6 +100,15 @@ enum
 #define DEFAULT_TARGET_RUNNING_TIME GST_CLOCK_TIME_NONE
 #define DEFAULT_MODE MODE_TIMECODE
 
+/* flags for self->must_send_end_message */
+enum
+{
+  END_MESSAGE_NORMAL = 0,
+  END_MESSAGE_STREAM_ENDED = 1,
+  END_MESSAGE_VIDEO_PUSHED = 2,
+  END_MESSAGE_AUDIO_PUSHED = 4
+};
+
 static void gst_avwait_set_property (GObject * object,
     guint prop_id, const GValue * value, GParamSpec * pspec);
 static void gst_avwait_get_property (GObject * object,
@@ -159,20 +176,24 @@ gst_avwait_class_init (GstAvWaitClass * klass)
       g_param_spec_boxed ("target-timecode", "Target timecode (object)",
           "Timecode to wait for in timecode mode (object)",
           GST_TYPE_VIDEO_TIME_CODE,
-          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+          GST_PARAM_MUTABLE_READY | G_PARAM_READWRITE |
+          G_PARAM_STATIC_STRINGS));
 
   g_object_class_install_property (gobject_class, PROP_TARGET_RUNNING_TIME,
       g_param_spec_uint64 ("target-running-time", "Target running time",
           "Running time to wait for in running-time mode",
           0, G_MAXUINT64,
           DEFAULT_TARGET_RUNNING_TIME,
-          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+          GST_PARAM_MUTABLE_READY | G_PARAM_READWRITE |
+          G_PARAM_STATIC_STRINGS));
 
   g_object_class_install_property (gobject_class, PROP_MODE,
       g_param_spec_enum ("mode", "Mode",
           "Operation mode: What to wait for",
           GST_TYPE_AVWAIT_MODE,
-          DEFAULT_MODE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+          DEFAULT_MODE,
+          GST_PARAM_MUTABLE_READY | G_PARAM_READWRITE |
+          G_PARAM_STATIC_STRINGS));
 
   g_object_class_install_property (gobject_class, PROP_END_TIME_CODE,
       g_param_spec_boxed ("end-timecode", "End timecode (object)",
@@ -180,6 +201,13 @@ gst_avwait_class_init (GstAvWaitClass * klass)
           GST_TYPE_VIDEO_TIME_CODE,
           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
+  g_object_class_install_property (gobject_class, PROP_RECORDING,
+      g_param_spec_boolean ("recording",
+          "Recording state",
+          "Whether the element is stopped or recording. "
+          "If set to FALSE, all buffers will be dropped regardless of settings.",
+          TRUE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
   gobject_class->finalize = gst_avwait_finalize;
   gstelement_class->change_state = gst_avwait_change_state;
 
@@ -243,15 +271,21 @@ gst_avwait_init (GstAvWait * self)
 
   self->running_time_to_wait_for = GST_CLOCK_TIME_NONE;
   self->last_seen_video_running_time = GST_CLOCK_TIME_NONE;
+  self->first_audio_running_time = GST_CLOCK_TIME_NONE;
   self->last_seen_tc = NULL;
 
   self->video_eos_flag = FALSE;
+  self->audio_eos_flag = FALSE;
+  self->video_flush_flag = FALSE;
   self->audio_flush_flag = FALSE;
   self->shutdown_flag = FALSE;
   self->dropping = TRUE;
   self->tc = gst_video_time_code_new_empty ();
   self->end_tc = NULL;
   self->running_time_to_end_at = GST_CLOCK_TIME_NONE;
+  self->audio_running_time_to_wait_for = GST_CLOCK_TIME_NONE;
+  self->audio_running_time_to_end_at = GST_CLOCK_TIME_NONE;
+  self->recording = TRUE;
 
   self->target_running_time = DEFAULT_TARGET_RUNNING_TIME;
   self->mode = DEFAULT_MODE;
@@ -259,6 +293,7 @@ gst_avwait_init (GstAvWait * self)
   gst_video_info_init (&self->vinfo);
   g_mutex_init (&self->mutex);
   g_cond_init (&self->cond);
+  g_cond_init (&self->audio_cond);
 }
 
 static void
@@ -271,6 +306,7 @@ gst_avwait_send_element_message (GstAvWait * self, gboolean dropping,
                   "dropping", G_TYPE_BOOLEAN, dropping,
                   "running-time", GST_TYPE_CLOCK_TIME, running_time, NULL)))) {
     GST_ERROR_OBJECT (self, "Unable to send element message!");
+    g_assert_not_reached ();
   }
 }
 
@@ -285,13 +321,17 @@ gst_avwait_change_state (GstElement * element, GstStateChange transition)
       g_mutex_lock (&self->mutex);
       self->shutdown_flag = TRUE;
       g_cond_signal (&self->cond);
+      g_cond_signal (&self->audio_cond);
       g_mutex_unlock (&self->mutex);
       break;
     case GST_STATE_CHANGE_READY_TO_PAUSED:
       g_mutex_lock (&self->mutex);
       self->shutdown_flag = FALSE;
       self->video_eos_flag = FALSE;
+      self->audio_eos_flag = FALSE;
+      self->video_flush_flag = FALSE;
       self->audio_flush_flag = FALSE;
+      self->must_send_end_message = END_MESSAGE_NORMAL;
       g_mutex_unlock (&self->mutex);
     default:
       break;
@@ -306,6 +346,8 @@ gst_avwait_change_state (GstElement * element, GstStateChange transition)
         GST_DEBUG_OBJECT (self, "First time reset in paused to ready");
         self->running_time_to_wait_for = GST_CLOCK_TIME_NONE;
         self->running_time_to_end_at = GST_CLOCK_TIME_NONE;
+        self->audio_running_time_to_wait_for = GST_CLOCK_TIME_NONE;
+        self->audio_running_time_to_end_at = GST_CLOCK_TIME_NONE;
       }
       if (!self->dropping) {
         self->dropping = TRUE;
@@ -317,6 +359,9 @@ gst_avwait_change_state (GstElement * element, GstStateChange transition)
       self->vsegment.position = GST_CLOCK_TIME_NONE;
       gst_video_info_init (&self->vinfo);
       self->last_seen_video_running_time = GST_CLOCK_TIME_NONE;
+      self->first_audio_running_time = GST_CLOCK_TIME_NONE;
+      if (self->last_seen_tc)
+        gst_video_time_code_free (self->last_seen_tc);
       self->last_seen_tc = NULL;
       g_mutex_unlock (&self->mutex);
       break;
@@ -344,6 +389,7 @@ gst_avwait_finalize (GObject * object)
 
   g_mutex_clear (&self->mutex);
   g_cond_clear (&self->cond);
+  g_cond_clear (&self->audio_cond);
 
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
@@ -374,6 +420,10 @@ gst_avwait_get_property (GObject * object, guint prop_id,
       g_value_set_uint64 (value, self->target_running_time);
       break;
     }
+    case PROP_RECORDING:{
+      g_value_set_boolean (value, self->recording);
+      break;
+    }
     case PROP_MODE:{
       g_value_set_enum (value, self->mode);
       break;
@@ -490,9 +540,11 @@ gst_avwait_set_property (GObject * object, guint prop_id,
       self->target_running_time = g_value_get_uint64 (value);
       if (self->mode == MODE_RUNNING_TIME) {
         self->running_time_to_wait_for = self->target_running_time;
+        if (self->recording) {
+          self->audio_running_time_to_wait_for = self->running_time_to_wait_for;
+        }
         if (self->target_running_time < self->last_seen_video_running_time) {
           self->dropping = TRUE;
-          gst_avwait_send_element_message (self, TRUE, GST_CLOCK_TIME_NONE);
         }
       }
       break;
@@ -508,14 +560,16 @@ gst_avwait_set_property (GObject * object, guint prop_id,
                     self->tc) < 0) {
               self->running_time_to_wait_for = GST_CLOCK_TIME_NONE;
               self->dropping = TRUE;
-              gst_avwait_send_element_message (self, TRUE, GST_CLOCK_TIME_NONE);
             }
             break;
           case MODE_RUNNING_TIME:
             self->running_time_to_wait_for = self->target_running_time;
+            if (self->recording) {
+              self->audio_running_time_to_wait_for =
+                  self->running_time_to_wait_for;
+            }
             if (self->target_running_time < self->last_seen_video_running_time) {
               self->dropping = TRUE;
-              gst_avwait_send_element_message (self, TRUE, GST_CLOCK_TIME_NONE);
             }
             break;
             /* Let the chain functions handle the rest */
@@ -527,6 +581,12 @@ gst_avwait_set_property (GObject * object, guint prop_id,
       }
       break;
     }
+    case PROP_RECORDING:{
+      g_mutex_lock (&self->mutex);
+      self->recording = g_value_get_boolean (value);
+      g_mutex_unlock (&self->mutex);
+      break;
+    }
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -553,6 +613,8 @@ gst_avwait_vsink_event (GstPad * pad, GstObject * parent, GstEvent * event)
         GST_DEBUG_OBJECT (self, "First time reset in video segment");
         self->running_time_to_wait_for = GST_CLOCK_TIME_NONE;
         self->running_time_to_end_at = GST_CLOCK_TIME_NONE;
+        self->audio_running_time_to_wait_for = GST_CLOCK_TIME_NONE;
+        self->audio_running_time_to_end_at = GST_CLOCK_TIME_NONE;
         if (!self->dropping) {
           self->dropping = TRUE;
           gst_avwait_send_element_message (self, TRUE, GST_CLOCK_TIME_NONE);
@@ -570,12 +632,21 @@ gst_avwait_vsink_event (GstPad * pad, GstObject * parent, GstEvent * event)
       g_cond_signal (&self->cond);
       g_mutex_unlock (&self->mutex);
       break;
+    case GST_EVENT_FLUSH_START:
+      g_mutex_lock (&self->mutex);
+      self->video_flush_flag = TRUE;
+      g_cond_signal (&self->audio_cond);
+      g_mutex_unlock (&self->mutex);
+      break;
     case GST_EVENT_FLUSH_STOP:
       g_mutex_lock (&self->mutex);
+      self->video_flush_flag = FALSE;
       if (self->mode != MODE_RUNNING_TIME) {
         GST_DEBUG_OBJECT (self, "First time reset in video flush");
         self->running_time_to_wait_for = GST_CLOCK_TIME_NONE;
         self->running_time_to_end_at = GST_CLOCK_TIME_NONE;
+        self->audio_running_time_to_wait_for = GST_CLOCK_TIME_NONE;
+        self->audio_running_time_to_end_at = GST_CLOCK_TIME_NONE;
         if (!self->dropping) {
           self->dropping = TRUE;
           gst_avwait_send_element_message (self, TRUE, GST_CLOCK_TIME_NONE);
@@ -636,6 +707,13 @@ gst_avwait_asink_event (GstPad * pad, GstObject * parent, GstEvent * event)
       g_cond_signal (&self->cond);
       g_mutex_unlock (&self->mutex);
       break;
+    case GST_EVENT_EOS:
+      g_mutex_lock (&self->mutex);
+      self->audio_eos_flag = TRUE;
+      self->must_send_end_message = END_MESSAGE_NORMAL;
+      g_cond_signal (&self->audio_cond);
+      g_mutex_unlock (&self->mutex);
+      break;
     case GST_EVENT_FLUSH_STOP:
       g_mutex_lock (&self->mutex);
       self->audio_flush_flag = FALSE;
@@ -670,6 +748,8 @@ gst_avwait_vsink_chain (GstPad * pad, GstObject * parent, GstBuffer * inbuf)
   GstClockTime running_time;
   GstVideoTimeCode *tc = NULL;
   GstVideoTimeCodeMeta *tc_meta;
+  gboolean retry = FALSE;
+  gboolean ret = GST_FLOW_OK;
 
   timestamp = GST_BUFFER_TIMESTAMP (inbuf);
   if (timestamp == GST_CLOCK_TIME_NONE) {
@@ -685,9 +765,24 @@ gst_avwait_vsink_chain (GstPad * pad, GstObject * parent, GstBuffer * inbuf)
 
   tc_meta = gst_buffer_get_video_time_code_meta (inbuf);
   if (tc_meta) {
-    tc = &tc_meta->tc;
+    tc = gst_video_time_code_copy (&tc_meta->tc);
+    if (self->last_seen_tc) {
+      gst_video_time_code_free (self->last_seen_tc);
+    }
     self->last_seen_tc = tc;
   }
+  while (self->mode == MODE_VIDEO_FIRST
+      && self->first_audio_running_time == GST_CLOCK_TIME_NONE
+      && !self->audio_eos_flag
+      && !self->shutdown_flag && !self->video_flush_flag) {
+    g_cond_wait (&self->audio_cond, &self->mutex);
+  }
+  if (self->video_flush_flag || self->shutdown_flag) {
+    GST_DEBUG_OBJECT (self, "Shutting down, ignoring buffer");
+    gst_buffer_unref (inbuf);
+    g_mutex_unlock (&self->mutex);
+    return GST_FLOW_FLUSHING;
+  }
   switch (self->mode) {
     case MODE_TIMECODE:{
       if (self->tc != NULL && tc != NULL) {
@@ -707,6 +802,10 @@ gst_avwait_vsink_chain (GstPad * pad, GstObject * parent, GstBuffer * inbuf)
           self->running_time_to_wait_for =
               gst_segment_to_running_time (&self->vsegment, GST_FORMAT_TIME,
               self->vsegment.position);
+          if (self->recording) {
+            self->audio_running_time_to_wait_for =
+                self->running_time_to_wait_for;
+          }
         }
         if (self->end_tc && gst_video_time_code_compare (tc, self->end_tc) >= 0) {
           if (self->running_time_to_end_at == GST_CLOCK_TIME_NONE) {
@@ -716,12 +815,14 @@ gst_avwait_vsink_chain (GstPad * pad, GstObject * parent, GstBuffer * inbuf)
             self->running_time_to_end_at =
                 gst_segment_to_running_time (&self->vsegment, GST_FORMAT_TIME,
                 self->vsegment.position);
-            gst_avwait_send_element_message (self, TRUE,
-                self->running_time_to_end_at);
+            if (self->recording) {
+              self->audio_running_time_to_end_at = self->running_time_to_end_at;
+              self->must_send_end_message |= END_MESSAGE_STREAM_ENDED;
+            }
           }
           gst_buffer_unref (inbuf);
           inbuf = NULL;
-        } else if (emit_passthrough_signal) {
+        } else if (emit_passthrough_signal && self->recording) {
           gst_avwait_send_element_message (self, FALSE,
               self->running_time_to_wait_for);
         }
@@ -739,7 +840,8 @@ gst_avwait_vsink_chain (GstPad * pad, GstObject * parent, GstBuffer * inbuf)
       } else {
         if (self->dropping) {
           self->dropping = FALSE;
-          gst_avwait_send_element_message (self, FALSE, running_time);
+          if (self->recording)
+            gst_avwait_send_element_message (self, FALSE, running_time);
         }
         GST_INFO_OBJECT (self,
             "Have %" GST_TIME_FORMAT ", waiting for %" GST_TIME_FORMAT,
@@ -755,21 +857,108 @@ gst_avwait_vsink_chain (GstPad * pad, GstObject * parent, GstBuffer * inbuf)
             self->vsegment.position);
         GST_DEBUG_OBJECT (self, "First video running time is %" GST_TIME_FORMAT,
             GST_TIME_ARGS (self->running_time_to_wait_for));
+        if (self->recording) {
+          self->audio_running_time_to_wait_for = self->running_time_to_wait_for;
+        }
         if (self->dropping) {
           self->dropping = FALSE;
-          gst_avwait_send_element_message (self, FALSE,
-              self->running_time_to_wait_for);
+          if (self->recording)
+            gst_avwait_send_element_message (self, FALSE,
+                self->running_time_to_wait_for);
         }
       }
       break;
     }
   }
+
+  if (!self->recording) {
+    if (self->was_recording) {
+      GST_INFO_OBJECT (self, "Recording stopped at %" GST_TIME_FORMAT,
+          GST_TIME_ARGS (running_time));
+      if (running_time > self->running_time_to_wait_for
+          && running_time <= self->running_time_to_end_at) {
+        /* We just stopped recording: synchronise the audio */
+        self->audio_running_time_to_end_at = running_time;
+        self->must_send_end_message |= END_MESSAGE_STREAM_ENDED;
+      } else if (running_time < self->running_time_to_wait_for
+          && self->running_time_to_wait_for != GST_CLOCK_TIME_NONE) {
+        self->audio_running_time_to_wait_for = GST_CLOCK_TIME_NONE;
+      }
+    }
+    /* Recording is FALSE: we drop all buffers */
+    if (inbuf) {
+      gst_buffer_unref (inbuf);
+      inbuf = NULL;
+    }
+  } else {
+    if (!self->was_recording) {
+      GST_INFO_OBJECT (self,
+          "Recording started at %" GST_TIME_FORMAT " waiting for %"
+          GST_TIME_FORMAT " inbuf %p", GST_TIME_ARGS (running_time),
+          GST_TIME_ARGS (self->running_time_to_wait_for), inbuf);
+      if (self->mode != MODE_VIDEO_FIRST ||
+          self->first_audio_running_time <= running_time ||
+          self->audio_eos_flag) {
+        if (running_time < self->running_time_to_end_at ||
+            self->running_time_to_end_at == GST_CLOCK_TIME_NONE) {
+          /* We are before the end of the recording. Check if we just actually
+           * started */
+          if (running_time > self->running_time_to_wait_for) {
+            /* We just started recording: synchronise the audio */
+            self->audio_running_time_to_wait_for = running_time;
+            gst_avwait_send_element_message (self, FALSE, running_time);
+          } else {
+            /* We will start in the future when running_time_to_wait_for is
+             * reached */
+            self->audio_running_time_to_wait_for =
+                self->running_time_to_wait_for;
+          }
+          self->audio_running_time_to_end_at = self->running_time_to_end_at;
+        }
+      } else {
+        /* We are in video-first mode and behind the first audio timestamp. We
+         * should drop all video buffers until the first audio timestamp, so
+         * we can catch up with it. (In timecode mode and running-time mode, we
+         * don't care about when the audio starts, we start as soon as the
+         * target timecode or running time has been reached) */
+        gst_buffer_unref (inbuf);
+        inbuf = NULL;
+        retry = TRUE;
+      }
+    }
+  }
+
+  if (!retry)
+    self->was_recording = self->recording;
   g_cond_signal (&self->cond);
   g_mutex_unlock (&self->mutex);
-  if (inbuf)
-    return gst_pad_push (self->vsrcpad, inbuf);
-  else
-    return GST_FLOW_OK;
+  if (inbuf) {
+    GST_WARNING_OBJECT (self, "Pass video buffer ending at %" GST_TIME_FORMAT,
+        GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (inbuf) +
+            GST_BUFFER_DURATION (inbuf)));
+    ret = gst_pad_push (self->vsrcpad, inbuf);
+  }
+  g_mutex_lock (&self->mutex);
+  if (self->must_send_end_message & END_MESSAGE_AUDIO_PUSHED) {
+    self->must_send_end_message = END_MESSAGE_NORMAL;
+    g_mutex_unlock (&self->mutex);
+    gst_avwait_send_element_message (self, TRUE,
+        self->audio_running_time_to_end_at);
+  } else if (self->must_send_end_message & END_MESSAGE_STREAM_ENDED) {
+    if (self->audio_eos_flag) {
+      self->must_send_end_message = END_MESSAGE_NORMAL;
+      g_mutex_unlock (&self->mutex);
+      gst_avwait_send_element_message (self, TRUE,
+          self->audio_running_time_to_end_at);
+    } else {
+      self->must_send_end_message |= END_MESSAGE_VIDEO_PUSHED;
+      g_mutex_unlock (&self->mutex);
+    }
+  } else {
+    g_mutex_unlock (&self->mutex);
+  }
+
+  return ret;
 }
 
 /*
@@ -800,6 +989,10 @@ gst_avwait_asink_chain (GstPad * pad, GstObject * parent, GstBuffer * inbuf)
   GstClockTime duration;
   GstClockTime running_time_at_end = GST_CLOCK_TIME_NONE;
   gint asign, vsign = 1, esign = 1;
+  GstFlowReturn ret = GST_FLOW_OK;
+  /* Make sure the video thread doesn't send the element message before we
+   * actually call gst_pad_push */
+  gboolean send_element_message = FALSE;
 
   timestamp = GST_BUFFER_TIMESTAMP (inbuf);
   if (timestamp == GST_CLOCK_TIME_NONE) {
@@ -817,6 +1010,10 @@ gst_avwait_asink_chain (GstPad * pad, GstObject * parent, GstBuffer * inbuf)
     GST_ERROR_OBJECT (self, "Could not get current running time");
     return GST_FLOW_ERROR;
   }
+  if (self->first_audio_running_time == GST_CLOCK_TIME_NONE) {
+    self->first_audio_running_time = current_running_time;
+  }
+  g_cond_signal (&self->audio_cond);
   if (self->vsegment.format == GST_FORMAT_TIME) {
     vsign =
         gst_segment_to_running_time_full (&self->vsegment, GST_FORMAT_TIME,
@@ -825,30 +1022,6 @@ gst_avwait_asink_chain (GstPad * pad, GstObject * parent, GstBuffer * inbuf)
       video_running_time = GST_CLOCK_TIME_NONE;
     }
   }
-  while (!(self->video_eos_flag || self->audio_flush_flag
-          || self->shutdown_flag) &&
-      /* Start at timecode */
-      /* Wait if we haven't received video yet */
-      (video_running_time == GST_CLOCK_TIME_NONE
-          /* Wait if audio is after the video: dunno what to do */
-          || gst_avwait_compare_guint64_with_signs (asign,
-              current_running_time, vsign, video_running_time) == 1
-          /* Wait if we don't even know what to wait for yet */
-          || self->running_time_to_wait_for == GST_CLOCK_TIME_NONE)) {
-    g_cond_wait (&self->cond, &self->mutex);
-    vsign =
-        gst_segment_to_running_time_full (&self->vsegment, GST_FORMAT_TIME,
-        self->vsegment.position, &video_running_time);
-    if (vsign == 0) {
-      video_running_time = GST_CLOCK_TIME_NONE;
-    }
-  }
-  if (self->audio_flush_flag || self->shutdown_flag) {
-    GST_DEBUG_OBJECT (self, "Shutting down, ignoring frame");
-    gst_buffer_unref (inbuf);
-    g_mutex_unlock (&self->mutex);
-    return GST_FLOW_FLUSHING;
-  }
   duration =
       gst_util_uint64_scale (gst_buffer_get_size (inbuf) / self->ainfo.bpf,
       GST_SECOND, self->ainfo.rate);
@@ -863,36 +1036,64 @@ gst_avwait_asink_chain (GstPad * pad, GstObject * parent, GstBuffer * inbuf)
       return GST_FLOW_ERROR;
     }
   }
-  if (self->running_time_to_wait_for == GST_CLOCK_TIME_NONE
+  while (!(self->video_eos_flag || self->audio_flush_flag
+          || self->shutdown_flag) &&
+      /* Start at timecode */
+      /* Wait if we haven't received video yet */
+      (video_running_time == GST_CLOCK_TIME_NONE
+          /* Wait if audio is after the video: dunno what to do */
+          || gst_avwait_compare_guint64_with_signs (asign,
+              running_time_at_end, vsign, video_running_time) == 1)) {
+    g_cond_wait (&self->cond, &self->mutex);
+    vsign =
+        gst_segment_to_running_time_full (&self->vsegment, GST_FORMAT_TIME,
+        self->vsegment.position, &video_running_time);
+    if (vsign == 0) {
+      video_running_time = GST_CLOCK_TIME_NONE;
+    }
+  }
+  if (self->audio_flush_flag || self->shutdown_flag) {
+    GST_DEBUG_OBJECT (self, "Shutting down, ignoring frame");
+    gst_buffer_unref (inbuf);
+    g_mutex_unlock (&self->mutex);
+    return GST_FLOW_FLUSHING;
+  }
+  if (self->audio_running_time_to_wait_for == GST_CLOCK_TIME_NONE
       /* Audio ends before start : drop */
       || gst_avwait_compare_guint64_with_signs (esign,
-          running_time_at_end, 1, self->running_time_to_wait_for) == -1
+          running_time_at_end, 1, self->audio_running_time_to_wait_for) == -1
       /* Audio starts after end: drop */
-      || current_running_time >= self->running_time_to_end_at) {
+      || current_running_time >= self->audio_running_time_to_end_at) {
     GST_DEBUG_OBJECT (self,
         "Dropped an audio buf at %" GST_TIME_FORMAT " waiting for %"
         GST_TIME_FORMAT " video time %" GST_TIME_FORMAT,
         GST_TIME_ARGS (current_running_time),
-        GST_TIME_ARGS (self->running_time_to_wait_for),
+        GST_TIME_ARGS (self->audio_running_time_to_wait_for),
         GST_TIME_ARGS (video_running_time));
     GST_DEBUG_OBJECT (self, "Would have ended at %i %" GST_TIME_FORMAT,
         esign, GST_TIME_ARGS (running_time_at_end));
     gst_buffer_unref (inbuf);
     inbuf = NULL;
+    if (current_running_time >= self->audio_running_time_to_end_at &&
+        (self->must_send_end_message & END_MESSAGE_STREAM_ENDED) &&
+        !(self->must_send_end_message & END_MESSAGE_AUDIO_PUSHED)) {
+      send_element_message = TRUE;
+    }
+
   } else if (gst_avwait_compare_guint64_with_signs (esign, running_time_at_end,
-          1, self->running_time_to_wait_for) >= 0
+          1, self->audio_running_time_to_wait_for) >= 0
       && gst_avwait_compare_guint64_with_signs (esign, running_time_at_end, 1,
-          self->running_time_to_end_at) == -1) {
+          self->audio_running_time_to_end_at) == -1) {
     /* Audio ends after start, but before end: clip */
     GstSegment asegment2 = self->asegment;
 
     gst_segment_set_running_time (&asegment2, GST_FORMAT_TIME,
-        self->running_time_to_wait_for);
+        self->audio_running_time_to_wait_for);
     inbuf =
         gst_audio_buffer_clip (inbuf, &asegment2, self->ainfo.rate,
         self->ainfo.bpf);
   } else if (gst_avwait_compare_guint64_with_signs (esign, running_time_at_end,
-          1, self->running_time_to_end_at) >= 0) {
+          1, self->audio_running_time_to_end_at) >= 0) {
     /* Audio starts after start, but before end: clip from the other side */
     GstSegment asegment2 = self->asegment;
     guint64 stop;
@@ -900,7 +1101,7 @@ gst_avwait_asink_chain (GstPad * pad, GstObject * parent, GstBuffer * inbuf)
 
     ssign =
         gst_segment_position_from_running_time_full (&asegment2,
-        GST_FORMAT_TIME, self->running_time_to_end_at, &stop);
+        GST_FORMAT_TIME, self->audio_running_time_to_end_at, &stop);
     if (ssign > 0) {
       asegment2.stop = stop;
     } else {
@@ -915,15 +1116,42 @@ gst_avwait_asink_chain (GstPad * pad, GstObject * parent, GstBuffer * inbuf)
     inbuf =
         gst_audio_buffer_clip (inbuf, &asegment2, self->ainfo.rate,
         self->ainfo.bpf);
+    if (self->must_send_end_message & END_MESSAGE_STREAM_ENDED) {
+      send_element_message = TRUE;
+    }
   } else {
     /* Programming error? Shouldn't happen */
     g_assert_not_reached ();
   }
   g_mutex_unlock (&self->mutex);
-  if (inbuf)
-    return gst_pad_push (self->asrcpad, inbuf);
-  else
-    return GST_FLOW_OK;
+  if (inbuf) {
+    GstClockTime new_duration =
+        gst_util_uint64_scale (gst_buffer_get_size (inbuf) / self->ainfo.bpf,
+        GST_SECOND, self->ainfo.rate);
+    GstClockTime new_running_time_at_end =
+        gst_segment_to_running_time (&self->asegment, GST_FORMAT_TIME,
+        self->asegment.position + new_duration);
+    GST_WARNING_OBJECT (self, "Pass audio buffer ending at %" GST_TIME_FORMAT,
+        GST_TIME_ARGS (new_running_time_at_end));
+    ret = gst_pad_push (self->asrcpad, inbuf);
+  }
+  if (send_element_message) {
+    g_mutex_lock (&self->mutex);
+    if ((self->must_send_end_message & END_MESSAGE_VIDEO_PUSHED) ||
+        self->video_eos_flag) {
+      self->must_send_end_message = END_MESSAGE_NORMAL;
+      g_mutex_unlock (&self->mutex);
+      gst_avwait_send_element_message (self, TRUE,
+          self->audio_running_time_to_end_at);
+    } else if (self->must_send_end_message & END_MESSAGE_STREAM_ENDED) {
+      self->must_send_end_message |= END_MESSAGE_AUDIO_PUSHED;
+      g_mutex_unlock (&self->mutex);
+    } else {
+      g_assert_not_reached ();
+    }
+  }
+  send_element_message = FALSE;
+  return ret;
 }
 
 static GstIterator *
