@@ -151,7 +151,7 @@ gst_wasapi_src_class_init (GstWasapiSrcClass * klass)
 
   gst_element_class_add_static_pad_template (gstelement_class, &src_template);
   gst_element_class_set_static_metadata (gstelement_class, "WasapiSrc",
-      "Source/Audio",
+      "Source/Audio/Hardware",
       "Stream audio from an audio capture device through WASAPI",
       "Nirbheek Chauhan <nirbheek@centricular.com>, "
       "Ole André Vadla Ravnås <ole.andre.ravnas@tandberg.com>");
